/**
 *
 * Base I/O device implementation.
 *
 * @file
 * Copyright &copy; Audi Electronics Venture GmbH. All rights reserved.
 *
 * $Author: ANOAFR4 $
 * $Date: 2014-06-27 16:33:30 +0200 (Fre, 27 Jun 2014) $
 * $Revision: 48654 $
 * @remarks
 *
 */
#include "stdafx.h"
#include <adtf_plugin_sdk.h>

namespace adtf
{
/**************************************************************************/
/* Private implementation to provide object compatibility                 */
/**************************************************************************/
class cBaseIODevice::cBaseIODevicePrivate : public cADTFD<cBaseIODevice, cBaseIODevice::cBaseIODevicePrivate>
{
    friend class cBaseIODevice;
    public:
        cBaseIODevicePrivate(){};
        virtual ~cBaseIODevicePrivate(){};
    protected:
        ucom::cReferenceCounter m_oReferenceCounter;
};

/**************************************************************************/
/* Base IO Device */
/**************************************************************************/
cBaseIODevice::cBaseIODevice()
{
    //d-pointer only used for reference counter at the moment
    ADTF_D_CREATE(cBaseIODevice);

    m_pEventHandler     = NULL;
}

cBaseIODevice::~cBaseIODevice()
{
    Close();
}

tResult cBaseIODevice::GetInterface(const tChar* idInterface, void ** ppvObject)
{
    if (ppvObject == NULL)
    {
        RETURN_ERROR(ERR_POINTER);
    }

    if (idmatch(idInterface, IID_DEVICE))
    {
        *ppvObject = static_cast<IDevice*> (this);
    }
    else if (idmatch(idInterface, IID_OBJECT))
    {
        *ppvObject = static_cast<IObject*> (static_cast<IDevice*> (this));
    }
    else
    {
        *ppvObject = NULL;
        RETURN_ERROR(ERR_NO_INTERFACE);
    }

    Ref();

    return ERR_NOERROR;
}

tUInt cBaseIODevice::Ref()
{
    return _d->m_oReferenceCounter.Inc();
}

tUInt cBaseIODevice::Unref()
{
    return _d->m_oReferenceCounter.Dec(this);
}

tVoid cBaseIODevice::Destroy()
{
    delete this;
}

tResult cBaseIODevice::Open(const tChar* strDevice, tInt nMode, __exception)
{
    return 0;
}

tResult cBaseIODevice::Close(__exception)
{
    return 0;
}

tInt cBaseIODevice::Read(tVoid* pvBuffer, tInt nBufferSize)
{
    return -1;
}

tInt cBaseIODevice::Write(const tVoid* pvBuffer, tInt nBufferSize)
{
    return -1;
}

tInt cBaseIODevice::IOCtl(tInt nCommand, tVoid* pvData, tInt nDataSize)
{
    switch (nCommand)
    {
        case IOCTL_GetDeviceInfo:
        {
            if (nDataSize >= (tInt)sizeof(tDeviceInfo))
            {
                tDeviceInfo* psDeviceInfo = (tDeviceInfo*) pvData;
                cMemoryBlock::MemSet(psDeviceInfo, 0, sizeof(tDeviceInfo));
                return sizeof(tDeviceInfo);
            }
            else
            {
                return -1;
            }
            break;
        }
        case IOCTL_RegisterHandler:
        {
            break;
        }
        case IOCTL_UnregisterHandler:
        {
            break;
        }
        default:
        {
            return -1;
        }
    }

    return 0;
}

tResult cBaseIODevice::SetDeviceEvent(tInt nEventId, tInt nParam1, tInt nParam2, tVoid* pvData, tInt nDataSize)
{
    if (m_pEventHandler != NULL)
    {
        return m_pEventHandler->OnDeviceEvent(nEventId, nParam1, nParam2, pvData, nDataSize);
    }

    RETURN_ERROR(ERR_NOT_SUPPORTED);
}

} // namespace adtf
